## Copyright (C) Solaar Contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from solaar.ui.about.model import AboutModel
from solaar.ui.about.presenter import Presenter
from solaar.ui.about.view import AboutView


def show(_=None, model=None, view=None):
    """Opens the About dialog."""
    if model is None:
        model = AboutModel()
    if view is None:
        view = AboutView()
    presenter = Presenter(model, view)
    presenter.run()


if __name__ == "__main__":
    from gi.repository import Gtk

    show(None)
    Gtk.main()
